
#include "IoTConnect.h"
WiFiClientSecure netMQ;
HTTPClient HttP;

    //// SDK info  ////
String BaseURL,CPID,VERG,UNIID,ENVT; 
typedef struct Sync_Resp{   
    String cpId="";
    String dtg="";
    int rc=0;               //root..info
    int ee=0;
    int at=0;
    int ds=0;
    struct ota{
      bool force=false;
      String guid="";
      //String urls[5];
      }OTA; 
    struct protocol{         
        String n="";
        String h="";
        String Id="";    //data..protocol
        String un="";
        String pwd="";
        String pT="";
        String sT="";
        } Broker;
};
struct Sync_Resp SYNC_resp;
String CMD_payload = "";
static void IOTConnectClientHandler(lwmqtt_client_t * /*client*/, void *ref, lwmqtt_string_t topic, lwmqtt_message_t message) {
    auto cb = (IOTConnectClientCallback *)ref;
    char terminated_topic[topic.len + 1];
    memcpy(terminated_topic, topic.data, topic.len);
    terminated_topic[topic.len] = '\0';
    if (message.payload != nullptr) {
        message.payload[message.payload_len] = '\0';
    }
    if (cb->advanced != nullptr) {
        cb->advanced(cb->client, terminated_topic, (char *)message.payload, (int)message.payload_len);
        return;
    }
    if (cb->simple == nullptr) {
        return;
    }
    String str_topic = String(terminated_topic);
    String str_payload;
    if (message.payload != nullptr) {
        str_payload = String((const char *)message.payload);
    }
    if (str_topic.substring(0, 30) == "$iothub/twin/PATCH/properties/"){
        DynamicJsonDocument root(512);  
        StaticJsonDocument<300> Twin_Json_in;
        deserializeJson(Twin_Json_in, str_payload);
        root["uniqueId"] = UNIID;
        root["desired"] = Twin_Json_in;
        str_payload = " ";
        serializeJson(root, str_payload);    
        cb->simpleT(str_topic, str_payload);
    }
    else if(str_topic.substring(0, 17) == "$iothub/twin/res/"){
        cb->simpleT(str_topic, str_payload);
    }
    else{
        StaticJsonDocument<512> incoming;  
        deserializeJson(incoming, str_payload);
        if( (incoming["cmdType"] == "0x01") ||(incoming["cmdType"] == "0x02")){
            String STR_Payload = incoming["data"].as<String>();
            cb->simple(str_topic, STR_Payload);
        }
        else CMD_payload = str_payload;
    }
} // End IOTConnectClientHandler 



void IOTConnectClient::begin(const char hostname[], Client &client) { this->begin(hostname, 1883, client); }
void IOTConnectClient::begin(const char hostname[], int port, Client &client) {
    this->setHost(hostname, port);
    this->netClient = &client;
    lwmqtt_init(&this->client, this->writeBuf, this->bufSize, this->readBuf, this->bufSize);
    lwmqtt_set_timers(&this->client, &this->timer1, &this->timer2, lwmqtt_arduino_timer_set, lwmqtt_arduino_timer_get);
    lwmqtt_set_network(&this->client, &this->network, lwmqtt_arduino_network_read, lwmqtt_arduino_network_write);
    lwmqtt_set_callback(&this->client, (void *)&this->callback, IOTConnectClientHandler);
}


void IOTConnectClient::onMessageAdvanced(IOTConnectClientCallbackAdvanced cb) {
    this->callback.client = this;
    this->callback.simple = nullptr;
    this->callback.advanced = cb;
}


void IOTConnectClient::setHost(const char hostname[]) { this->setHost(hostname, 1883); }
void IOTConnectClient::setHost(const char hostname[], int port) {
    if (this->hostname != nullptr) {
        free((void *)this->hostname);
    }
    this->hostname = strdup(hostname);
    this->port = port;
}


void IOTConnectClient::setWill(const char topic[]) { this->setWill(topic, ""); }
void IOTConnectClient::setWill(const char topic[], const char payload[]) { this->setWill(topic, payload, false, 0); }
void IOTConnectClient::setWill(const char topic[], const char payload[], bool retained, int qos) {
    if (topic == nullptr || strlen(topic) == 0) {
        return;
    }
    this->clearWill();
    this->will = (lwmqtt_will_t *)malloc(sizeof(lwmqtt_will_t));
    memset(this->will, 0, sizeof(lwmqtt_will_t));
    this->will->topic = lwmqtt_string(strdup(topic));
    if (payload != nullptr && strlen(payload) > 0) {
        this->will->payload = lwmqtt_string(strdup(payload));
    }
    this->will->retained = retained;
    this->will->qos = (lwmqtt_qos_t)qos;
}


void IOTConnectClient::clearWill() {
    if (this->will == nullptr) {
        return;
    }
    if (this->will->payload.len > 0) {
        free(this->will->payload.data);
    }
    if (this->will->topic.len > 0) {
      free(this->will->topic.data);
    }
    free(this->will);
    this->will = nullptr;
}

  
void IOTConnectClient::setOptions(int keepAlive, bool cleanSession, int timeout) {
    this->keepAlive = (uint16_t)keepAlive;
    this->cleanSession = cleanSession;
    this->timeout = (uint32_t)timeout;
}


bool IOTConnectClient::connect(const char clientId[]) { return this->connect(clientId, nullptr, nullptr); }
bool IOTConnectClient::connect(const char clientId[], const char username[]) { return this->connect(clientId, username, nullptr); }
bool IOTConnectClient::connect(const char clientId[], const char username[], const char password[]) {
    bool skip = false;
    if (!skip && this->connected()) {
        this->close();
    }
    this->network.client = this->netClient;
    if(!skip) {
        int ret = this->netClient->connect(this->hostname, (uint16_t)this->port);
        if (ret <= 0) {
            return false;
        }
    }
    lwmqtt_options_t options = lwmqtt_default_options;
    options.keep_alive = this->keepAlive;
    options.clean_session = this->cleanSession;
    options.client_id = lwmqtt_string(clientId);
    if (username != nullptr) {
        options.username = lwmqtt_string(username);
        if (password != nullptr) {
            options.password = lwmqtt_string(password);
        }
    }
    this->_lastError = lwmqtt_connect(&this->client, options, this->will, &this->_returnCode, this->timeout);
    if (this->_lastError != LWMQTT_SUCCESS) {
        this->close();
        return false;
    }
    this->_connected = true;
    return true;
}


bool IOTConnectClient::publish(const String &topic) { return this->publish(topic.c_str(), ""); }
bool IOTConnectClient::publish(const char topic[]) { return this->publish(topic, ""); }
bool IOTConnectClient::publish(const String &topic, const String &payload) { return this->publish(topic.c_str(), payload.c_str()); }
bool IOTConnectClient::publish(const String &topic, const String &payload, bool retained, int qos) {
    return this->publish(topic.c_str(), payload.c_str(), retained, qos);    }
bool IOTConnectClient::publish(const char topic[], const String &payload) { return this->publish(topic, payload.c_str()); }
bool IOTConnectClient::publish(const char topic[], const String &payload, bool retained, int qos) {
    return this->publish(topic, payload.c_str(), retained, qos);    }
bool IOTConnectClient::publish(const char topic[], const char payload[]) {
    return this->publish(topic, (char *)payload, (int)strlen(payload));   }
bool IOTConnectClient::publish(const char topic[], const char payload[], bool retained, int qos) {
    return this->publish(topic, (char *)payload, (int)strlen(payload), retained, qos);    }
bool IOTConnectClient::publish(const char topic[], const char payload[], int length) {
    return this->publish(topic, payload, length, false, 0);   }
bool IOTConnectClient::publish(const char topic[], const char payload[], int length, bool retained, int qos) {
    if (!this->connected()) {
        return false;
    }
    lwmqtt_message_t message = lwmqtt_default_message;
    message.payload = (uint8_t *)payload;
    message.payload_len = (size_t)length;
    message.retained = retained;
    message.qos = lwmqtt_qos_t(qos);
    this->_lastError = lwmqtt_publish(&this->client, lwmqtt_string(topic), message, this->timeout);
    if (this->_lastError != LWMQTT_SUCCESS) {
        this->close();
        return false;
    }

    return true;
}


bool IOTConnectClient::subscribe(const String &topic) { return this->subscribe(topic.c_str()); }
bool IOTConnectClient::subscribe(const String &topic, int qos) { return this->subscribe(topic.c_str(), qos); }
bool IOTConnectClient::subscribe(const char topic[]) { return this->subscribe(topic, 0); }
bool IOTConnectClient::subscribe(const char topic[], int qos) {
    if (!this->connected()) {
        return false;
    }
    this->_lastError = lwmqtt_subscribe_one(&this->client, lwmqtt_string(topic), (lwmqtt_qos_t)qos, this->timeout);
    if (this->_lastError != LWMQTT_SUCCESS) {
        this->close();
        return false;
    }
    return true;
}


bool IOTConnectClient::unsubscribe(const String &topic) { return this->unsubscribe(topic.c_str()); }
bool IOTConnectClient::unsubscribe(const char topic[]) {
    if (!this->connected()) {
        return false;
    }
    this->_lastError = lwmqtt_unsubscribe_one(&this->client, lwmqtt_string(topic), this->timeout);
    if (this->_lastError != LWMQTT_SUCCESS) {
        this->close();
        return false;
    }
    return true;
}

bool IOTConnectClient::MQTT_loop() {
    if (!this->connected()) {
        return false;
    }
    auto available = (size_t)this->netClient->available();
    if (available > 0) {
        this->_lastError = lwmqtt_yield(&this->client, available, this->timeout);
        if (this->_lastError != LWMQTT_SUCCESS) {
            this->close();
            return false;
        }
    }
    Received_cmd(CMD_payload);
    this->_lastError = lwmqtt_keep_alive(&this->client, this->timeout);
    if (this->_lastError != LWMQTT_SUCCESS) {
        this->close();
        return false;
    }
    return true;
}


bool IOTConnectClient::connected() {
    return this->netClient != nullptr && this->netClient->connected() == 1 && this->_connected;
}

  
bool IOTConnectClient::disconnect() {
    if (!this->connected()) {
        return false;
    }
    this->_lastError = lwmqtt_disconnect(&this->client, this->timeout);
    this->close();
    return this->_lastError == LWMQTT_SUCCESS;
}


void IOTConnectClient::InitializationSDK(String incpid, String UniID, IOTConnectClientCallbackSimple CallBack, IOTConnectClientCallbackSimple TwinCallBack, String env){
    String SyncResponseData;
    CPID = incpid;
    UNIID = UniID;
    ENVT = env;
    VERG = "2.0";
    if(Flag_99){
    if ((WiFi.status() == WL_CONNECTED)) { //Check the current connection status
        netMQ.setFingerprint(FPrint); // SHA1 Fingerprint
        netMQ.setTimeout(15000); // 15 Seconds
        delay(1000);
        if (!netMQ.connect(discovery, 443)) {
              Serial.println("discovery connection failed");
              return;
              }                             
          String dis_url = "/api/sdk/cpid/"+CPID+"/lang/M_C/ver/"+VERG+"/env/"+ENVT ; 
          netMQ.print(String("GET ") + dis_url + " HTTP/1.1\r\n" +
               "Host: " + discovery + "\r\nConnection: close\r\n\r\n");
          delay(50);    
          String payload = "";
          boolean httpBody = false;
          while (netMQ.connected()) {
              String line = netMQ.readStringUntil('\r');
              if (!httpBody && line.charAt(1) == '{') {
                  httpBody = true;
                  }
              if (httpBody) {
                  payload += line;
                  }
          }
          if( payload.length() >0) {
              StaticJsonDocument<256> root;
              deserializeJson(root, payload);
              const char* getBase = root["baseUrl"];  
              BaseURL = (String)getBase + "sync";
              Serial.println("\nYour Environment > " + ENVT);
              Serial.println("Your SDK Lang > M_C" );
              Serial.println("Your SDK Version > " + VERG);
              Serial.println("Base_URL > " + BaseURL); 
          }     // till now we get baseurl and now time to sync responce    
      SyncResponseData = SyncCall(CPID, UNIID, BaseURL);
      }
        SaveSyncResponce(SyncResponseData);
        MQTTConnection();
    
        // CallBack function
        this->callback.client = this;
        this->callback.simple = CallBack;
        this->callback.simpleT = TwinCallBack;
        this->callback.advanced = nullptr;      
  }
}


String IOTConnectClient::SyncCall(String incpid, String UniID, String BaseURL){
    String PayLoad ="";
    String postData = "{\"cpId\":\""+incpid+"\",\"uniqueId\":\""+UniID+"\",\"option\":{\"attribute\":false,\"setting\":false,\"protocol\":true,\"device\":false,\"sdkConfig\":false,\"rule\":false}}";
    if(WiFi.status()== WL_CONNECTED){   //Check WiFi connection status
        String type = BaseURL.substring(0,7);
        if(type == "http://"){
            hr:
            HttP.begin(BaseURL);      //Specify request destination
            HttP.addHeader("Accept", "application/json");  //Specify content-type header
            HttP.addHeader("Content-Type", "application/json");   
            int httpCode = HttP.POST(postData);   //Send the request
            PayLoad = HttP.getString();                  //Get the response payload
            if(httpCode >0);
            else 
                goto hr;
        }
        else if(type == "https:/"){
            String comm =  BaseURL.substring(8);
            char *command = new char[comm.length()+1];
            strcpy(command,comm.c_str());
            const char *hoSt = strtok(command,"/");
            String URl = BaseURL.substring(8 + strlen(hoSt));
            if ((WiFi.status() == WL_CONNECTED)) { //Check the current connection status
                netMQ.setFingerprint(FPrint);
                netMQ.setTimeout(15000); // 15 Seconds
                delay(1000);
            
                if (!netMQ.connect(hoSt, 443)) {
                    Serial.println("connection failed with Host");
                    return "Host not connected";
                }  
                netMQ.print(String("POST ") + URl + " HTTP/1.1\r\n" +
                    "Host: "+hoSt+"\r\n"+"Connection: close\r\n"+"Accept: */*\r\n"+
                    "Content-Type: application/json;charset=utf-8\r\n" +
                    "Content-Length: "+postData.length()+"\r\n" +
                    "\r\n" + postData + "\r\n"+
                    "Connection: close\r\n\r\n");
                boolean httpBody = false;
                while (netMQ.connected()) {
                    String line = netMQ.readStringUntil('\r');
                    if (!httpBody && line.charAt(1) == '{')
                        httpBody = true;
                    if (httpBody)
                        PayLoad += line;
                    }
            }else;
    }}
    return PayLoad;
}


void IOTConnectClient::SaveSyncResponce(String SyncResponseData){
    DynamicJsonDocument Sync_Res_Json(2048);
//    Serial.println(SyncResponseData);  
    DeserializationError error = deserializeJson(Sync_Res_Json, SyncResponseData); 
    if (!error){
        SYNC_resp.cpId = Sync_Res_Json["d"]["cpId"].as<String>();;
        SYNC_resp.dtg = Sync_Res_Json["d"]["dtg"].as<String>();
        SYNC_resp.ee = Sync_Res_Json["d"]["ee"].as<int>();
        SYNC_resp.rc = Sync_Res_Json["d"]["rc"].as<int>();
        SYNC_resp.at = Sync_Res_Json["d"]["at"].as<int>();
        SYNC_resp.ds = Sync_Res_Json["d"]["ds"].as<int>();
        SYNC_resp.OTA.force = Sync_Res_Json["d"]["ota"]["force"].as<bool>();    
        SYNC_resp.OTA.guid = Sync_Res_Json["d"]["ota"]["guid"].as<String>();   
        SYNC_resp.Broker.n = Sync_Res_Json["d"]["p"]["n"].as<String>();       
        SYNC_resp.Broker.h = Sync_Res_Json["d"]["p"]["h"].as<String>();    
        SYNC_resp.Broker.Id = Sync_Res_Json["d"]["p"]["id"].as<String>();   
        SYNC_resp.Broker.un = Sync_Res_Json["d"]["p"]["un"].as<String>();    
        SYNC_resp.Broker.pwd = Sync_Res_Json["d"]["p"]["pwd"].as<String>(); 
        SYNC_resp.Broker.pT = Sync_Res_Json["d"]["p"]["pub"].as<String>();   
        SYNC_resp.Broker.sT = Sync_Res_Json["d"]["p"]["sub"].as<String>(); 
      Serial.println("\tSync_Response_Data Saved") ;
      
      if(SYNC_resp.ds == 0)
          Serial.println(" \t.. OK ");
      else if(SYNC_resp.ds == 1)
          Serial.println(" \tDevice_Not_Register ");        
      else if(SYNC_resp.ds == 2)
          Serial.println(" \tAuto_Register ");
      else if(SYNC_resp.ds == 3)
          Serial.println(" \tDevice_Not_Found ");
      else if(SYNC_resp.ds == 4)
          Serial.println(" \tDevice_Inactive ");
      else if(SYNC_resp.ds == 5)
          Serial.println(" \tObject_Moved ");
      else if(SYNC_resp.ds == 6)
          Serial.println(" \tCpid_Not_Found ");
      else
          Serial.println("No Device_status has been matched..! 000");
    }
    else Serial.println("SyncResponse have no data Object") ;  //End if strcmp(new_object,"data")
    delay(100);

    Sync_Res_Json.clear();
}  //End save_data_SRensponse


void IOTConnectClient::MQTTConnection(){  
//========================================================================================================
  char* const FPrint2 = "bb 24 b6 56 6a 96 91 79 ee 86 6b aa 94 31 3d ab 7e 24 61 17";
//========================================================================================================
    while (WiFi.status() != WL_CONNECTED) {
        Serial.print(".");
        delay(100);
        }
    Serial.println("\tConnected.  WiFI ") ;  
    netMQ.setFingerprint(FPrint);  
    begin(string2char(SYNC_resp.Broker.h), 8883, netMQ);
    delay(50);
    while (!connect(string2char(SYNC_resp.Broker.Id), string2char(SYNC_resp.Broker.un), string2char(SYNC_resp.Broker.pwd) )) {
        Serial.print(".");
        delay(800);
    }
    Serial.println("\tMQTT connected") ;
    subscribe(string2char(SYNC_resp.Broker.sT),  1);
    subscribe(string2char(twinPropertySubTopic), 0);
    subscribe(string2char(twinResponseSubTopic), 0);
} //End MQTTConnection

void IOTConnectClient::GetAllTwins(void){
  delay(100);
  if (!connected())
      MQTTConnection();
  if(publish(string2char(twinResponsePubTopic), " ", 0))
        Serial.println("\t^^ Twin_Response_Publish ^^");
}

void IOTConnectClient::HTTPConnection(String incoming){
    String PayLoad ="";
    Serial.println(SYNC_resp.Broker.h);
    String urlX= "https://"+SYNC_resp.Broker.h+"/devices/"+SYNC_resp.Broker.Id+"/messages/events?api-version="+httpAPIVersion;
    Serial.println(urlX);
    if(WiFi.status()== WL_CONNECTED){   //Check WiFi connection status
        netMQ.setFingerprint(FPrint);
        netMQ.setTimeout(15000); // 15 Seconds
        delay(100);
        if (!netMQ.connect(SYNC_resp.Broker.h, 443)) {
            Serial.println("connection failed with Host");
            }            
        netMQ.print(String("POST ") + urlX + " HTTP/1.1\r\n" +
        "Host: "+SYNC_resp.Broker.h+"\r\n"+"Connection: close\r\n"+"Accept: */*\r\n"+
        "Content-Type: application/json;charset=utf-8\r\n" +
        "Content-Length: "+incoming.length()+"\r\n" +
        "\r\n" + incoming + "\r\n");
        //"Connection: close\r\n\r\n");   
    }
    Serial.println("\t\tHTTPS published\n\n");
}   //END of HTTPConnection

// this function will get data for IoTConnect
void IOTConnectClient::SendData(String Attribute_json_Data){
  if(Flag_99){
  if(!SYNC_resp.ds){
    String in_time[10],in_ID[10];
    JsonObject in_Attr[10];
    delay(50); 
    DynamicJsonDocument Attribute_json(2048);
    deserializeJson(Attribute_json, Attribute_json_Data);
    JsonArray root = Attribute_json.as<JsonArray>();    
    int Asize = root.size();
    for(int x = 0; x < Asize ;x++){    
        JsonObject row = root[x].as<JsonObject>();;
        in_ID[x] = row["uniqueId"].as<String>();
        in_time[x] = row["time"].as<String>();
        in_Attr[x] = row["data"].as<JsonObject>();
    } 

    DynamicJsonDocument To_HUB_json(2048);
    To_HUB_json["cpId"] = SYNC_resp.cpId;
    To_HUB_json["dtg"] = SYNC_resp.dtg;
    To_HUB_json["t"] = in_time[0];
    To_HUB_json["mt"] = 0;
    JsonObject sdk_info = To_HUB_json.createNestedObject("sdk");
       sdk_info["l"] = "M_C";
       sdk_info["v"] = VERG;
       sdk_info["e"] = ENVT;
    JsonArray device = To_HUB_json.createNestedArray("d");
    for(int x = 0; x < Asize ;x++){
        JsonObject Device = device.createNestedObject();
        Device["id"] = in_ID[x];
        Device["tg"] = "";
        Device["dt"] = in_time[x];
            JsonArray Attributes = Device.createNestedArray("d");
            Attributes.add(in_Attr[x]);
    }
    String To_HUB_json_Data;
    serializeJson(To_HUB_json, To_HUB_json_Data);
    Serial.println(To_HUB_json_Data);
    sendToAzure("Device_Attributes_Data", To_HUB_json_Data);
    delay(100);
    To_HUB_json.clear();
  }
 }
} //End SendDataINput


//This will UpdateTwin property to IoTConnect
void IOTConnectClient::UpdateTwin(String key,String value){
   String Twin_Json_Data;
   DynamicJsonDocument root(256);  
   root[key] = value;
   serializeJson(root, Twin_Json_Data);
   Serial.println(Twin_Json_Data);
   root.clear(); 
   if (!connected())
      MQTTConnection();
   if(publish(string2char(twinPropertyPubTopic), Twin_Json_Data, 0, 1))
      Serial.println("\t^^ UpdateTwin Published ^^");
}


// When you received command for SDK process
void IOTConnectClient::Received_cmd(String payload){
    String PayLoad="",Cmd_value="";  
    StaticJsonDocument<512> incmd;   
    DeserializationError error = deserializeJson(incmd, payload);
    if (!error){
        if(incmd["cmdType"].as<String>() == "0x12"){
            disconnect();                delay(10000);
            PayLoad = SyncCall(CPID, UNIID, BaseURL);
            SaveSyncResponce(PayLoad); MQTTConnection();
            }
        else if(incmd["cmdType"].as<String>() == "0x10")
            Serial.println("Get Response code for Attribute Changed");        
        else if(incmd["cmdType"].as<String>() == "0x11")
            Serial.println("Get Response code for Setting Changed");
        else if(incmd["cmdType"].as<String>() == "0x13")
            Serial.println("Get Response code for Device Changed");
        else if(incmd["cmdType"].as<String>() == "0x15")
            Serial.println("Get Response code for Rule Changed");
        else if(incmd["cmdType"].as<String>() == "0x99"){
            disconnect(); Flag_99 = false;   
            }
        else
            Serial.println("No response code has been matched..! 000");
    }
}//End Received_cmd()


char* IOTConnectClient::string2char(String command){
    if(command.length()!=0){
        char *p = const_cast<char*>(command.c_str());
        return p;
    }
}


// sendToAzure send data to HUB by .... Protocol
int IOTConnectClient::sendToAzure(String call_from, String pubData){
      if(SYNC_resp.Broker.n == "http" ||SYNC_resp.Broker.n == "https"){
          Serial.println("Your Protocol >>  ..HTTPS.. ");
          HTTPConnection(pubData);
          }
      else if(SYNC_resp.Broker.n == "mqtt" ||SYNC_resp.Broker.n == "mqtts"){
          Serial.println("Your Protocol >>  ..MQTTS.. ");
          if (!connected())
              MQTTConnection();
          if(publish(string2char(SYNC_resp.Broker.pT), pubData , 0, 1)){
              delay(10);
              
              Serial.print("\t^^ "+ call_from);
              Serial.println("\tPublished ^^");
              }
          }
      else
          Serial.println("Your Protocol >>  ..Unknown.. ");
}

bool IOTConnectClient::Dispose(){
    if (disconnect()){
		Flag_99 = false; 
        Serial.println("\n\t:: SDK is Disconnected From IoTConnect ::");
    }
}


// this will send the ACK of receiving Commands
void IOTConnectClient::SendAck(String Ack_Data, int messageType){
    StaticJsonDocument<350> Ack_Json;
    Ack_Json["uniqueId"] = UNIID;
    Ack_Json["cpId"] = CPID;
    Ack_Json["mt"] = messageType;
    JsonObject sdk_info = Ack_Json.createNestedObject("sdk");
       sdk_info["l"] = "M_C";
       sdk_info["v"] = VERG;
       sdk_info["e"] = ENVT;
    StaticJsonDocument<300> Ack_Json_in;
    deserializeJson(Ack_Json_in, Ack_Data);
    Ack_Json["d"]= Ack_Json_in;
 
    String Ack_Json_Data;
    serializeJson(Ack_Json, Ack_Json_Data);
    sendToAzure("Ack_Data", Ack_Json_Data);
    Ack_Json.clear();    Ack_Json_in.clear();  
}
